//
//  AnnotationScanner.swift
//  swiftgraph
//
//  Created by Karl Traunmüller on 15.10.19.
//  Copyright © 2019 Runtastic. All rights reserved.
//

import Foundation

import SourceKittenFramework

public final class AnnotationScanner {
    
    private let componentMarker = "#@component"
    private let configurationMarker = "#@configuration"
    
    public init() {
    }
    
    public func scanFolder(at path: String) throws -> DependencyInfo? {
        let fileManager = FileManager()
        guard let enumerator = fileManager.enumerator(atPath: path) else {
            return nil
        }
        let codeBaseDependencyInfo = DependencyInfo(components: [], configurations: [])
        while let relativePath = enumerator.nextObject() as? String {
            guard (relativePath as NSString).pathExtension == "swift" else {
                continue
            }
            print(relativePath)
            let filePath = (path as NSString).appendingPathComponent(relativePath)
            let fileContents = try String(contentsOfFile: filePath, encoding: .utf8)
            if let fileDependencyInfo = try scanFile(fileContents: fileContents) {
                codeBaseDependencyInfo.merge(with: fileDependencyInfo)
            }
        }
        return codeBaseDependencyInfo
    }
    
    public func scanFile(fileContents: String) throws -> DependencyInfo? {
        let dependencies = DependencyInfo(components: [], configurations: [])
        let file = File(contents: fileContents)
        let structure = try Structure(file: file)
        //print(toNSDictionary(structure.dictionary))
        let substructures = structure.dictionary[SwiftDocKey.substructure.rawValue] as? [[String: SourceKitRepresentable]] ?? []
        for substructure in substructures {
            guard let kindRawValue = substructure[SwiftDocKey.kind.rawValue] as? String,
                let kind = SwiftDeclarationKind(rawValue: kindRawValue),
                let name = substructure[SwiftDocKey.name.rawValue] as? String else {
                    continue
            }
            switch kind {
            case .class:
                print("scanning class \(name)...")
            case .struct:
                print("scanning struct \(name)...")
            case .extension:
                print("scanning extension \(name)...")
            default:
                continue
            }
            guard let docBlock = extractDocumentationBlock(fileContents: fileContents, substructure: substructure) else {
                continue
            }
            var components: [Component] = []
            var configurations: [Configuration] = []
            let lines = (docBlock as NSString).lines()
            for line in lines {
                if line.content.contains(componentMarker) {
                    print("\(kind) \(name) is a component!")
                    if let component = scanComponent(substructure) {
                        components.append(component)
                    }
                } else if line.content.contains(configurationMarker) {
                    print("\(kind) \(name) is a configuration!")
                    if let configuration = scanConfiguration(substructure) {
                        configurations.append(configuration)
                    }
                }
            }
            dependencies.merge(with: DependencyInfo(components: components, configurations: configurations))
        }
        
        return dependencies
    }
    
    private func extractDocumentationBlock(fileContents: String, substructure: [String: SourceKitRepresentable]) -> String? {
        guard let docOffset = substructure["key.docoffset"] as? Int64,
            let docLength = substructure["key.doclength"] as? Int64 else {
                return nil
        }
        return (fileContents as NSString).substringLinesWithByteRange(start: Int(docOffset), length: Int(docLength))
    }
    
    private func scanComponent(_ substructure: [String: SourceKitRepresentable]) -> Component? {
        guard let name = substructure[SwiftDocKey.name.rawValue] as? String else {
            return nil
        }
        let (properties, warnings) = scanProperties(typeName: name, substructure)
        return Component(typeName: name, properties: properties, warnings: warnings)
    }
    
    private func scanConfiguration(_ substructure: [String: SourceKitRepresentable]) -> Configuration? {
        guard let name = substructure[SwiftDocKey.name.rawValue] as? String else {
            return nil
        }
        let (properties, warnings) = scanProperties(typeName: name, substructure)
        return Configuration(typeName: name, properties: properties, warnings: warnings)
    }
    
    private func scanProperties(typeName structureTypeName: String, _ substructure: [String: SourceKitRepresentable]) -> (Set<Property>, warnings: [String]) {
        var properties: Set<Property> = []
        var warnings: [String] = []
        let members = substructure[SwiftDocKey.substructure.rawValue] as? [[String: SourceKitRepresentable]] ?? []
        for member in members {
            guard let kindRawValue = member[SwiftDocKey.kind.rawValue] as? String,
                let kind = SwiftDeclarationKind(rawValue: kindRawValue),
                let name = member[SwiftDocKey.name.rawValue] as? String else {
                    continue
            }
            switch kind {
            case .varInstance, .varStatic, .varClass:
                guard let typeName = member[SwiftDocKey.typeName.rawValue] as? String else {
                    warnings.append("Missing type information for property \(name) on \(structureTypeName)")
                    continue
                }
                let property = Property(name: name,
                                        typeName: typeName.trimmingCharacters(in: CharacterSet(charactersIn: "?")),
                                        isStatic: kind == .varStatic || kind == .varClass,
                                        isOptional: typeName.contains("?"),
                                        isWeak: hasAttribute(member, "source.decl.attribute.weak"))
                properties.insert(property)
            default:
                break
            }
        }
        return (properties, warnings)
    }
    
    private func hasAttribute(_ structure: [String: SourceKitRepresentable], _ attribute: String) -> Bool {
        guard let attributes = structure["key.attributes"] as? [[String: SourceKitRepresentable]] else {
            return false
        }
        return attributes.contains(where: { $0.values.contains(where: { $0 as? String == attribute }) })
    }
}
