//
//  Models.swift
//  swiftgraph
//
//  Created by Karl Traunmüller on 16.10.19.
//  Copyright © 2019 Runtastic. All rights reserved.
//

public struct Property: Hashable {
    let name: String
    let typeName: String
    let isStatic: Bool
    let isOptional: Bool
    let isWeak: Bool
}

public final class Component {
    let typeName: String
    
    /// this is the entire list of properties;
    /// these are then matched against discovered components and configurations
    private(set) var properties: Set<Property>
    
    let warnings: [String]
    
    init(typeName: String, properties: Set<Property>, warnings: [String]) {
        self.typeName = typeName
        self.properties = properties
        self.warnings = warnings
    }
    
    func merge(with other: Component) {
        properties.formUnion(other.properties)
    }
    
    func firstPropertyOfType(_ typeName: String) -> Property? {
        return properties.first(where: { $0.typeName == typeName })
    }
}

public final class Configuration {
    let typeName: String
    
    /// this is the entire list of properties;
    /// these are then matched against discovered components and configurations
    private(set) var properties: Set<Property>
    
    let warnings: [String]
    
    init(typeName: String, properties: Set<Property>, warnings: [String]) {
        self.typeName = typeName
        self.properties = properties
        self.warnings = warnings
    }
    
    func merge(with other: Configuration) {
        properties.formUnion(other.properties)
    }
}

public final class DependencyInfo {
    
    public private(set) var components: [Component]
    public private(set) var configurations: [Configuration]
        
    public var warnings: [String] {
        let configurationWarnings = configurations.reduce([], { $0 + $1.warnings })
        let componentWarnings = components.reduce([], { $0 + $1.warnings })
        return configurationWarnings + componentWarnings
    }
    
    init(components: [Component], configurations: [Configuration]) {
        self.components = components
        self.configurations = configurations
    }
    
    public func firstConfigurationOfType(_ typeName: String) -> Configuration? {
        return configurations.first(where: { $0.typeName == typeName })
    }

    public func firstComponentOfType(_ typeName: String) -> Component? {
        return components.first(where: { $0.typeName == typeName })
    }

    func merge(with other: DependencyInfo) {
        if components.isEmpty {
            components = other.components
        } else {
            for component in other.components {
                if let existingComponent = self.components.first(where: { $0.typeName == component.typeName }) {
                    existingComponent.merge(with: component)
                } else {
                    self.components.append(component)
                }
            }
        }
        if configurations.isEmpty {
            configurations = other.configurations
        } else {
            for configuration in other.configurations {
                if let existingConfiguration = self.configurations.first(where: { $0.typeName == configuration.typeName }) {
                    existingConfiguration.merge(with: configuration)
                } else {
                    self.configurations.append(configuration)
                }
            }
        }
    }
}

public struct DependencyGraphEdge: Hashable {
    
    enum Kind {
        case strongComponentReference
        case weakComponentReference
        case configurationReference
        case ruleViolation
    }
    
    let fromTypeName: String
    let toTypeName: String
    let kind: Kind
}
